import time
import argparse
from pathlib import Path

from .distance import distance
from .geoindex import GeoIndex
from .table import Table


def cli():

    note1 = '---------------------------------------------------------------'
    note2 = 'Notes:'
    note3 = '- The --distance and --neigbor arguments can be repeated.'
    note4 = '\nExample:\n> tupu some_cities.csv \n\t-o augmented.csv \n\t-d dist_ny,40.71,-74.0 \n\t-d dist_dc,38.9,-77.0 \n\t-n dist_reserve,id_reserve,reserve_cities.tsv?id=feature_id\n\t-n dist_self,id_self'
    notes = '\n'.join((note1, note2, note3, note4, note1))

    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter, description='tupu: calculate geodesic distances between points and sets', epilog=notes)
    parser.add_argument('filename')
    parser.add_argument("--verbose", help="increase output verbosity", action="store_true")
    parser.add_argument("--timeit", help="report elapsed time", action="store_true")
    parser.add_argument('--output','-o', action='store', help='output filename; comma-separated for .csv files, else tab-separated', required=True)
    parser.add_argument('--distance','-d', action='append', help='compute distance to point <lat,lon> and store it in column <dist>: "dist,lat,lon"')
    parser.add_argument('--neighbor','-n', action='append', help='compute distance to closest point in <fn>, and store it in columns <dist,id>: "dist,id,fn". If <fn> is omitted, will compare against self (input filename)')
    args = parser.parse_args()

    verbose = args.verbose
    timeit = args.timeit
    if timeit: start_time = time.time()
    if verbose: print(' - Validating input...')

    # Parse input filename and load input data
    input_table = Table(args.filename, verbose=verbose)

    # Check that the output file has the correct extension
    out_fn = Path(args.output)
    assert out_fn.suffix in ('.raw', '.tsv', '.csv'), out_fn.suffix

    # Compute distances between each input row and given points
    for _ in args.distance:
        dist_name, lat, lon = validate_cli_distance(_)
        if verbose:
            print(f' - Computing distance to lat={lat} lon={lon} and storing it in "{dist_name}"')
            input_table.add_distance_to_point(lat, lon, dist_name)

    # Compute minimum distances between each input row and a given *set* of points
    for _ in args.neighbor:
        id_name, dist_name, fn = validate_cli_neighbor(_)

        if fn is None:
            # Distance to the list itself
            input_table.add_distance_to_self(id_name, dist_name)
        else:
            # Distance to another list
            neighbor_table = Table(fn, verbose=verbose)
            input_table.add_distance_to_table(neighbor_table, id_name, dist_name)

    # Save data        
    input_table.save(out_fn)
    
    # Done!
    if verbose:
        print('\nDone!')
    if timeit:
        print("--- {:7.3f} seconds ---".format(time.time() - start_time))


def validate_cli_distance(input):
    input = input.split(',')
    assert len(input) == 3, f'--distance requires a (name,lat,lon) list,  but received {input}'
    name, lat, lon = input
    lat = float(lat)
    lon = float(lon)
    assert -90 <= lat <= 90
    assert -180 <= lon <= 180
    return name, lat, lon


def validate_cli_neighbor(input):
    input = input.split(',')
    assert len(input) in (2,3), f'--neighbor requires either <id,dist,fn> or <id,dist>, but received {input}'
    if len(input) == 2:
        input.append(None)
    id_name, dist_name, fn = input
    return id_name, dist_name, fn



"""
Distance wrapper around pyproj
"""


import pyproj

GEOD = pyproj.Geod(ellps='WGS84')

def distance(lat1, lon1, lat2, lon2):
    """
    Return distance between two points in km
    """
    d = GEOD.inv(lon1, lat1, lon2, lat2)[2] / 1000
    return d



from pathlib import Path

from .utils import load_csv, save_csv, get_args_from_filename, get_coords
from .geoindex import GeoIndex
from .distance import distance

class Table:

    def __init__(self, filename, verbose=False):

        self.verbose = verbose

        # Parse filename (extract optional arguments, check that the file exists)
        fn, self.args = get_args_from_filename(filename)
        fn = Path(fn)
        assert fn.exists(), fn
        self.filename = Path(fn)

        self.header, self.data = load_csv(self.filename)
        if verbose:
            print(' - Filename:', str(self.filename))
            print(' - Headers:', self.header)
            print(' - Number of rows:', len(self.data))

        self.lat_i = self.header.index(self.args['lat'])
        self.lon_i = self.header.index(self.args['lon'])
        self.id_i = self.header.index(self.args['id'])

        self._index = None

    @property
    def index(self):
        # Create index if needed
        if self._index is None:
            coords = get_coords(self.header, self.data, self.args)
            self._index = GeoIndex(data=coords)
        return self._index


    def add_distance_to_point(self, lat, lon, name):
        assert name not in self.header
        self.header.append(name)

        verbose = self.verbose
        lat_i = self.lat_i
        lon_i = self.lon_i

        for i, row in enumerate(self.data, 1):
            if verbose and (i % 1000 == 0):
                print('.', end='', flush=True)
            row_lat, row_lon = float(row[lat_i]), float(row[lon_i])
            dist = distance(row_lat, row_lon, lat, lon)
            row.append(f'{dist:.4f}')
        if verbose and i >= 1000:
            print()


    def add_distance_to_table(self, other_table, id_name, dist_name):

        if self.verbose:
            print(f' - Computing distance to nearest neighbor from {other_table.filename}; storing distance in "{dist_name}" and neighbor id in "{id_name}"')
        
        assert id_name not in self.header
        assert dist_name not in self.header
        self.header.extend([id_name, dist_name])

        # Create index
        idx = other_table.index

        verbose = self.verbose
        lat_i = self.lat_i
        lon_i = self.lon_i

        # Compute all distances
        for i, row in enumerate(self.data, 1):
            if verbose and (i % 1000 == 0):
                print('.', end='', flush=True)
            
            lat = float(row[lat_i])
            lon = float(row[lon_i])
            
            neighbor_id = idx.nearest(lat, lon)
            dist = distance(lat, lon, *idx.coords[neighbor_id])
            row.extend([f'{dist:.4f}', neighbor_id])

        if verbose and i >= 1000:
            print()


    def add_distance_to_self(self, id_name, dist_name):

        if self.verbose:
            print(f' - Computing distance to nearest neighbor from self; storing distance in "{dist_name}" and neighbor id in "{id_name}"')

        assert id_name not in self.header
        assert dist_name not in self.header
        self.header.extend([id_name, dist_name])

        # Create index
        idx = self.index

        verbose = self.verbose
        lat_i = self.lat_i
        lon_i = self.lon_i
        id_i = self.id_i

        # Compute all distances
        for i, row in enumerate(self.data, 1):
            if verbose and (i % 1000 == 0):
                print('.', end='', flush=True)
            
            lat = float(row[lat_i])
            lon = float(row[lon_i])
            identifier = int(row[id_i])

            neighbor_id = idx.nearest(lat, lon, identifier=identifier)
            dist = distance(lat, lon, *idx.coords[neighbor_id])
            row.extend([f'{dist:.4f}', neighbor_id])

        if verbose and i >= 1000:
            print()


    def save(self, filename):
        if self.verbose:
            print(f' - Saving output in {filename}')
        delimiter = ',' if filename.suffix == '.csv' else '\t'
        save_csv(filename, self.header, self.data, delimiter=delimiter)


"""
tupu: fast geodesic distances
================================
tupu is a Python package that calculates
`geodesic distances<https://en.wikipedia.org/wiki/Geodesics_on_an_ellipsoid>`_
between points or sets of points.

Its goal is to be able to efficiently compute distances to lists of points.
For instance, if you have a list of N points and M cities, and want to compute
the closest city to each point, it avoids brute-force computation of the NxM
distances, and instead indexes each city through
`R-Trees<https://en.wikipedia.org/wiki/R-tree>`_.

tupu is build on two key packages:

1. Geodesic distances are computed through pyproj, a Python interface to the PROJ.4 library (written in C).
2. Spatial indexing is done through the `rtree<http://toblerity.org/rtree/>`_ package,
a Python wrapper to the `libspatialindex<https://libspatialindex.org/>`_ library.
"""


from .utils import load_csv, save_csv, get_args_from_filename, get_coords
from .distance import distance
from .geoindex import GeoIndex
from .table import Table
from .io import cli



import csv


def load_csv(fn):
    ext = fn.suffix
    delim = ',' if ext == '.csv' else '\t'
    with open(fn) as fh:
        reader = csv.reader(fh, delimiter=delim)
        header = next(reader)
        data = [row for row in reader]
        return header, data


def save_csv(fn, header, data, delimiter):
    with open(fn, 'w', newline='') as fh:
        writer = csv.writer(fh, delimiter=delimiter, quoting=csv.QUOTE_MINIMAL)
        writer.writerow(header)
        writer.writerows(data)


def get_args_from_filename(fn):
    # Allow arguments that change default header names
    # Example: some_fn.csv?lat=latitude&lon=long
    default_args = {'lat':'lat', 'lon':'lon', 'id':'id'}
    if '?' in fn:
        fn, new_args = fn.split('?', maxsplit=1)
        default_args.update(kv.split('=') for kv in new_args.split('&'))
    return fn, default_args


def get_coords(header, data, args):
    lat_i = header.index(args['lat'])
    lon_i = header.index(args['lon'])
    id_i = header.index(args['id'])
    #print(header)
    #print(id_i, lat_i, lon_i)
    coords = [(int(row[id_i]), float(row[lat_i]), float(row[lon_i])) for row in data]
    return coords




"""
Spatial index using an R-Tree
"""


# ---------------------------
# Imports
# ---------------------------

import pyproj
import rtree
from .distance import distance

# ---------------------------
# Classes
# ---------------------------

class GeoIndex:

    def __init__(self, data=None, leaf_capacity=100, near_minimum_overlap_factor=32):
        """ <data> is an iterator containing (id, x, y) elements.
        User must ensure that id is unique!
        """

        self.coords = dict()    # id -> (lat, lon)
        self.xy = dict()        # id -> (x, y)

        # API: https://pyproj4.github.io/pyproj/html/api/proj.html#pyproj-proj
        # Projection: Albers Equal Area, https://proj4.org/operations/projections/aea.html
        self.proj = pyproj.Proj('+proj=aea +lat_1=29.5 +lat_2=42.5')

        # Reference: https://libspatialindex.org/overview.html#the-rtree-package
        p = rtree.index.Property()
        p.leaf_capacity = leaf_capacity
        p.near_minimum_overlap_factor = near_minimum_overlap_factor

        if data is None:
            self.idx = rtree.index.Index(properties=p)
        else:
            # stream takes an iterable of the form (id, (x,y,x,y), object) where object is MANDATORY but can be none
            data_generator = rtree_generator(self.proj, self.coords, self.xy, data)
            self.idx = rtree.index.Index(data_generator, properties=p)
            assert len(self.coords), 'Index is empty even though a stream was provided'


    def insert(self, identifier, lat, lon):
        assert type(identifier) == int
        assert type(lat) == type(lon) == float
        assert identifier not in self.coords
        x, y = self.proj(lon, lat)
        
        self.coords[identifier] = (lat, lon)
        self.xy[identifier] = (x, y)
        self.idx.insert(identifier, (x, y, x, y) )


    def delete(self, identifier):
        self.coords.pop(identifier)
        x, y = self.xy.pop(identifier)
        self.idx.delete(identifier, (x, y, x, y) )


    def nearest(self, lat, lon, identifier=None, num_results=1):
        """Return the N elements closest to a (lat, lon) point.
        - If identifier is not None, we will exclude answers with that id (which we take as "self")
        - Each element is returned as (ID, distance)
        """
        
        #assert num_results==1
        # TODO: Allow nearest N (need to get the Nth and then use that as pivot)
        # TODO: Allow nearest by group?? but that probably is multiple indexes
        # see original rtreegeo.py code for more sanity checks

        x, y = self.proj(lon, lat)
        _ = (x, y, x, y)
        candidates = [c for c in self.idx.nearest(_, num_results + 1) if c != identifier]
        return candidates[0]


# ---------------------------
# Functions
# ---------------------------

def rtree_generator(proj, coords, xy, data):
    for identifier, lat, lon in data:
        assert type(identifier) == int
        assert type(lat) in (float, int) and type(lon) in (float, int), (lat, lon)
        assert identifier not in coords
        x, y = proj(lon, lat)
        coords[identifier] = (lat, lon)
        xy[identifier] = (x, y)
        yield (identifier, (x, y, x, y), None)
